#include <string.h>
#include "globdef.h"
#include "conf.h"
#include "uidef.h"
#include "padef.h"
#include "thrdef.h"
#include "sigdef.h"
#include "keyboard_def.h"
#include "fft1def.h"
#include "screendef.h"

#define MAX_LATENCY 20

#if PORTAUDIO_PRESENT == 1
PaStream *out_stream;
PaStream *in_stream;
#if(OSNUM == OS_FLAG_LINUX)
#if PA_LINUX_REALTIME == 1
void PaAlsa_EnableRealtimeScheduling( PaStream *s, int enable );
#endif
#if JACK_PRESENT == 1
const char linrad_name[8] ="linrad";
PaError PaJack_SetClientName(const char *name);

#endif
#endif

int daout_workload_reset;
/// This routine will be called by the PortAudio engine when audio is needed.
/// It may be called at interrupt level on some machines so don't do anything
/// that could mess up the system like calling malloc() or free().

extern int recordCallback( const void *inputBuffer, void *outputBuffer,
                         unsigned long framesToProcess,
                         const PaStreamCallbackTimeInfo* timeInfo,
                         PaStreamCallbackFlags statusFlags,
                         void *userData )
{
// Prevent unused variable warnings.
(void) outputBuffer;
(void) timeInfo;
(void) statusFlags;
(void) userData;
if(portaudio_active_flag == FALSE)
  {
  lirerr(701401);
  return paAbort;
  }
if(snd[RXAD].block_frames != (int)framesToProcess)
  {
  lirerr(520674);
  return paAbort;
  }
memcpy(&timf1_char[timf1p_sdr],inputBuffer,snd[RXAD].block_bytes);
timf1p_sdr=(timf1p_sdr+snd[RXAD].block_bytes)&timf1_bytemask;
lir_set_event(EVENT_PORTAUDIO_RXREADY);
return paContinue;
}

extern int playCallback( const void *inputBuffer, void *outputBuffer,
                         unsigned long framesToProcess,
                         const PaStreamCallbackTimeInfo* timeInfo,
                         PaStreamCallbackFlags statusFlags,
                         void *userData )
{
// Prevent unused variable warnings.
(void) inputBuffer;
(void) timeInfo;
(void) statusFlags;
(void) userData;
int samps;
if(portaudio_active_flag == FALSE)
  {
  lirerr(701400);
  return paAbort;
  }
if(snd[RXDA].block_frames != (int)framesToProcess)
  {
  lirerr(520673);
  return paAbort;
  }
if(thread_command_flag[THREAD_RX_OUTPUT]==THRFLAG_IDLE ||
           thread_status_flag[THREAD_RX_OUTPUT]==THRFLAG_IDLE)goto idle;
if(daout_workload_reset!=workload_reset_flag)
  {
  min_daout_samps=baseband_size;
  daout_workload_reset=workload_reset_flag;
  }
if(audio_dump_flag == 1)goto idle;
samps=(daout_pa-daout_px+daout_size)&daout_bufmask;
if( samps > snd[RXDA].block_bytes)
  {
  memcpy(outputBuffer,&daout[daout_px],snd[RXDA].block_bytes);
  daout_px=(daout_px+snd[RXDA].block_bytes)&daout_bufmask;
  samps-=snd[RXDA].block_bytes;
  if(min_daout_samps > samps)min_daout_samps=samps;
  }
else
  {
  min_daout_samps=0;
idle:;  
  memset(outputBuffer, 0, snd[RXDA].block_bytes);
  }
return paContinue;
}

void pa_rx_input(void)
{
while (!kill_all_flag && timf1p_sdr == timf1p_pa)
  {
  lir_await_event(EVENT_PORTAUDIO_RXREADY);
  }
}


void portaudio_stop(void)
{
if(portaudio_active_flag == FALSE)
  {
  lirerr(701401);
  return;
  }
Pa_Terminate();
}

void open_portaudio_rxad(void)
{
PaStreamParameters  rxadParameters;
const PaStreamInfo *streamInfo;
PaError pa_err;
if(portaudio_active_flag == FALSE)
  {
  lir_rxin_status=LIR_PA_FLAG_ERROR;
  lirerr(701402);
  return;
  }
// Set sample format
if( (ui.rx_input_mode&DWORD_INPUT) == 0)
  {
  rxadParameters.sampleFormat = paInt16; // 2 byte samples
  }
else
  {  
  rxadParameters.sampleFormat = paInt32; // 4 byte samples
  }
rxadParameters.device       = ui.rx_addev_no;
rxadParameters.channelCount = ui.rx_ad_channels;
rxadParameters.suggestedLatency = ui.rx_ad_latency/snd[RXAD].interrupt_rate;
// ******************************************************************************
#if (OSNUM == OS_FLAG_WINDOWS)
//TO GET BEST LATENCY WITH WASAPI: SELECT EXCLUSIVE MODE INSTEAD OF THE DEFAULT SHARED MODE 
//EXCLUSIVE MODE WORKS FOR VISTAX64, VISTAX86, WINDOWS7
//EXCLUSIVE MODE DOES NOT WORK FOR WOW64 (32-bit applications running on VISTAX64)
const PaDeviceInfo *deviceInfo;
int i;
deviceInfo = Pa_GetDeviceInfo(ui.rx_addev_no);
i = strncmp(Pa_GetHostApiInfo( deviceInfo->hostApi )->name,"Windows WASAPI", 14);
if (i==0)
  {
  PaWasapiStreamInfo LinradWasapiStreaminfo;
  LinradWasapiStreaminfo.size=sizeof(PaWasapiStreamInfo);
  LinradWasapiStreaminfo.hostApiType=paWASAPI;
  LinradWasapiStreaminfo.version=1;
  LinradWasapiStreaminfo.flags=paWinWasapiExclusive;
  rxadParameters.hostApiSpecificStreamInfo = &LinradWasapiStreaminfo;
  }
else
#endif
  {
  rxadParameters.hostApiSpecificStreamInfo = NULL;
  }
// ********************************************************************************
pa_err = Pa_IsFormatSupported( &rxadParameters, NULL, ui.rx_ad_speed);
if( pa_err != paFormatIsSupported )
  {
  DEB"\npa_err=%s\n",Pa_GetErrorText(pa_err));
  lir_rxin_status=LIR_PA_FORMAT_NOT_SUPPORTED;
  goto start_pa_end;
  }
timf1p_sdr=timf1p_pa;  
lir_init_event(EVENT_PORTAUDIO_RXREADY);
pa_err = Pa_IsFormatSupported( &rxadParameters, NULL, ui.rx_ad_speed);
// Open and Start audio I/O stream.
pa_err = Pa_OpenStream( 
              &in_stream,
              &rxadParameters,
              NULL,
              (double)(ui.rx_ad_speed),           //SAMPLE_RATE,
              snd[RXAD].block_frames,             //frames per buffer
              paClipOff,                          //no clipping
              recordCallback,
              NULL );                             //no userdata
if( pa_err != paNoError )
  {
  DEB"\npa_err=%s\n",Pa_GetErrorText(pa_err));
  lir_close_event(EVENT_PORTAUDIO_RXREADY);
  lir_rxin_status=LIR_PA_OPENSTREAM_FAILED;
  goto start_pa_end;
  }
#if PA_LINUX_REALTIME == 1
#if ALSA_PRESENT == 1
PaAlsa_EnableRealtimeScheduling ( in_stream, 1);
#endif
#endif
pa_err = Pa_StartStream( in_stream );
if( pa_err != paNoError )
  {
  DEB"\npa_err=%s\n",Pa_GetErrorText(pa_err));
  lir_rxin_status=LIR_PA_START_STREAM_FAILED;
  goto start_pa_end;
  }
streamInfo = Pa_GetStreamInfo(in_stream);
DEB"\nLATENCY %f\n",streamInfo->inputLatency);
DEB"\nRXout interrupt_rate: Actual %f",snd[RXAD].interrupt_rate);
start_pa_end:;
snd[RXAD].tot_bytes=snd[RXAD].no_of_blocks*snd[RXAD].block_bytes;
}

void open_portaudio_rxda(void)
{
PaStreamParameters  rxdaParameters;
const PaStreamInfo *streamInfo;
PaError pa_err;
if(portaudio_active_flag == FALSE)
  {
  lir_rxout_status=LIR_PA_FLAG_ERROR;
  lirerr(701401);
  return;
  }
// Set sample format
if(rx_daout_bytes == 1)
  {
  rxdaParameters.sampleFormat = paUInt8; // 1 byte samples
  }
else
  {  
  rxdaParameters.sampleFormat = paInt16; // 2 byte samples
  }
rxdaParameters.device       = ui.rx_dadev_no;
rxdaParameters.channelCount = rx_daout_channels;
rxdaParameters.suggestedLatency = ui.rx_da_latency/snd[RXDA].interrupt_rate;
// ******************************************************************************
#if (OSNUM == OS_FLAG_WINDOWS)
//TO GET BEST LATENCY WITH WASAPI: SELECT EXCLUSIVE MODE INSTEAD OF THE DEFAULT SHARED MODE 
//EXCLUSIVE MODE WORKS FOR VISTAX64, VISTAX86, WINDOWS7
//EXCLUSIVE MODE DOES NOT WORK FOR WOW64 (32-bit applications running on VISTAX64)
const PaDeviceInfo *deviceInfo;
int i;
deviceInfo = Pa_GetDeviceInfo(ui.rx_dadev_no);
i = strncmp(Pa_GetHostApiInfo( deviceInfo->hostApi )->name,"Windows WASAPI", 14);
if (i==0)
  {
  PaWasapiStreamInfo LinradWasapiStreaminfo;
  LinradWasapiStreaminfo.size=sizeof(PaWasapiStreamInfo);
  LinradWasapiStreaminfo.hostApiType=paWASAPI;
  LinradWasapiStreaminfo.version=1;
  LinradWasapiStreaminfo.flags=paWinWasapiExclusive;
  rxdaParameters.hostApiSpecificStreamInfo = &LinradWasapiStreaminfo;
  }
else
#endif
  {
  rxdaParameters.hostApiSpecificStreamInfo = NULL;
  }
// ********************************************************************************
pa_err = Pa_IsFormatSupported( NULL,&rxdaParameters, genparm[DA_OUTPUT_SPEED] );
if( pa_err != paFormatIsSupported )
  {
  lir_rxout_status=LIR_PA_FORMAT_NOT_SUPPORTED;
  goto start_pa_end;
  }
// Open and Start audio I/O stream.
pa_err = Pa_OpenStream( 
              &out_stream,
              NULL,                               //no input
              &rxdaParameters,
              (double)(genparm[DA_OUTPUT_SPEED]), //SAMPLE_RATE,
              snd[RXDA].block_frames,             //frames per buffer
              paClipOff,                          //no clipping
              playCallback,
              NULL );                             //no userdata
if( pa_err != paNoError )
  {
  lir_rxout_status=LIR_PA_OPENSTREAM_FAILED;
  goto start_pa_end;
  }

#if PA_LINUX_REALTIME == 1
#if ALSA_PRESENT == 1
PaAlsa_EnableRealtimeScheduling ( out_stream, 1);
#endif
#endif
pa_err = Pa_StartStream( out_stream );
if( pa_err != paNoError )
  {
  lir_rxout_status=LIR_PA_START_STREAM_FAILED;
  goto start_pa_end;
  }
streamInfo = Pa_GetStreamInfo(out_stream);
DEB"\nLATENCY %f\n",streamInfo->outputLatency);
DEB"\nRXout interrupt_rate: Actual %f",snd[RXDA].interrupt_rate);
start_pa_end:;
// set an arbitrary positive value.
if(lir_rxout_status == LIR_OK)rx_audio_out=1;
snd[RXDA].tot_bytes=snd[RXDA].no_of_blocks*snd[RXDA].block_bytes;
daout_workload_reset=workload_reset_flag;
}

void close_portaudio_rxda(void)
{
PaError pa_err;
pa_err = Pa_CloseStream( out_stream );
if( pa_err != paNoError )
  {
  lirerr(1168);
  }
}

void close_portaudio_rxad(void)
{
PaError pa_err;
pa_err = Pa_CloseStream( in_stream );
lir_close_event(EVENT_PORTAUDIO_RXREADY);
if( pa_err != paNoError )
  {
  lirerr(1229);
  }
}

int pa_get_device_info (int  n,
                        int sound_type,
                        void *pa_device_name,
                        void *pa_device_hostapi, 
			double *pa_device_max_speed,
			double *pa_device_min_speed,
			int *pa_device_max_bytes,
			int *pa_device_min_bytes,
			int *pa_device_max_channels,
			int *pa_device_min_channels )

{
(void) n ;
(void) sound_type;
(void) pa_device_name;
(void) pa_device_hostapi; 
(void) pa_device_max_speed;
(void) pa_device_min_speed;
(void) pa_device_max_bytes;
(void) pa_device_min_bytes;
(void) pa_device_max_channels;
(void) pa_device_min_channels;
const PaDeviceInfo *deviceInfo;
PaError pa_err;
PaStreamParameters inputParameters;
PaStreamParameters outputParameters;
int i,j, speed_warning;
int minBytes, maxBytes;
int maxInputChannels, maxOutputChannels;
int minInputChannels, minOutputChannels;
double maxStandardSampleRate;
double minStandardSampleRate;
// negative terminated  list
static double standardSampleRates[] = {8000.0, 9600.0, 
        11025.0, 12000.0, 16000.0, 22050.0, 24000.0, 32000.0,
        44100.0, 48000.0, 88200.0, 96000.0, 192000.0, -1};
// *******************************************************
sprintf((char*)(pa_device_name),"%s"," ");
sprintf((char*)(pa_device_hostapi),"%s"," ");
*pa_device_max_speed=0;
*pa_device_min_speed=0;
*pa_device_max_bytes=0;
*pa_device_min_bytes=0;
*pa_device_max_channels=0;
*pa_device_min_channels=0;
minInputChannels=0;
minOutputChannels=0;
if(n >= Pa_GetDeviceCount() ) return -1;
deviceInfo = Pa_GetDeviceInfo(n);
if (deviceInfo->maxOutputChannels==0 && 
         (sound_type==RXDA || sound_type==TXDA))return -1;
if (deviceInfo->maxInputChannels==0 && 
         (sound_type==RXAD || sound_type==TXAD))return -1; 
sprintf((char*)(pa_device_name),"%s",deviceInfo->name);
sprintf((char*)(pa_device_hostapi),"%s",Pa_GetHostApiInfo( deviceInfo->hostApi )->name);
speed_warning=0;
#if (OSNUM == OS_FLAG_WINDOWS)
// bypass bug in Juli@ ASIO driver: 
// this driver hangs after a Pa_IsFormatSupported call 
i = strncmp(deviceInfo->name, "ASIO 2.0 - ESI Juli@", 19);
if (i == 0)
    {
    minStandardSampleRate=44100;
    maxStandardSampleRate=192000;
    minBytes=1;
    maxBytes=4;
    maxInputChannels= deviceInfo->maxInputChannels;
    minInputChannels= 1;
    maxOutputChannels= deviceInfo->maxOutputChannels;
    minOutputChannels= 1;
    goto end_pa_get_device_info;
    }
#endif
// Investigate device capabilities.
// Check min and max samplerates  with 16 bit data.
maxStandardSampleRate=0;
minStandardSampleRate=0;
inputParameters.device = n;
inputParameters.channelCount = deviceInfo->maxInputChannels;
inputParameters.sampleFormat = paInt16;
inputParameters.suggestedLatency = 0;
inputParameters.hostApiSpecificStreamInfo = NULL;
outputParameters.device = n;
outputParameters.channelCount = deviceInfo->maxOutputChannels;
outputParameters.sampleFormat = paInt16; /*paInt24;*/
outputParameters.suggestedLatency = 0;
outputParameters.hostApiSpecificStreamInfo = NULL;
SNDLOG"\n\nTESTING PORTAUDIO DEVICE %i FOR %s\n",n, sndtype[sound_type]);
SNDLOG"\nid=%d device=%s hostapi=%s max_input_chan=%d max_output_chan=%d",
          n, 
          deviceInfo->name,
          Pa_GetHostApiInfo( deviceInfo->hostApi )->name,
          deviceInfo->maxInputChannels,
          deviceInfo->maxOutputChannels);
fflush(sndlog); 
// ************************************************************************
//filter for portaudio Windows hostapi's with non experts.
#if (OSNUM == OS_FLAG_WINDOWS)
//only allow ASIO or WASAPI or WDM-KS
i = strncmp(Pa_GetHostApiInfo(deviceInfo->hostApi)->name, "ASIO", 4);
if (i==0 ) goto end_filter_hostapi;
i = strncmp(Pa_GetHostApiInfo(deviceInfo->hostApi)->name, "Windows WASAPI", 14);
if (i==0 ) goto end_filter_hostapi;
i = strncmp(Pa_GetHostApiInfo(deviceInfo->hostApi)->name, "Windows WDM-KS", 14);
if (i==0 ) goto end_filter_hostapi;
if(ui.operator_skil == OPERATOR_SKIL_EXPERT)
  {
  speed_warning=1;
  }
else  
  { 
  return -1;
  }
end_filter_hostapi:;
#endif
// ************************************************************************
i=0;
while(standardSampleRates[i] > 0 && minStandardSampleRate==0)
   {
    SNDLOG"\nTesting %d %f",i,standardSampleRates[i]);
    if(sound_type==RXDA || sound_type==TXDA)
      {
      pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, standardSampleRates[i] );
      }
    else
      {  
      pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, standardSampleRates[i] );
      }
    SNDLOG"  err=%d  %s",pa_err, Pa_GetErrorText(pa_err));
    fflush(sndlog);  
    if(pa_err == paDeviceUnavailable)return -1;
    if(pa_err == paInvalidDevice)    return -1;
    if(pa_err == paFormatIsSupported )
      { 
      minStandardSampleRate=standardSampleRates[i];
      }
    i++;
   }
if(minStandardSampleRate == 0) return -1;
j=i;
while(standardSampleRates[i] > 0 )i++;
i--;
while(i >= j && maxStandardSampleRate==0)
   {
    SNDLOG"\nTesting %d %f",i,standardSampleRates[i]);
    if(sound_type==RXDA || sound_type==TXDA)
      {
      pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, standardSampleRates[i] );
      }
    else
      {  
      pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, standardSampleRates[i] );
      }
    SNDLOG"  err=%d  %s",pa_err, Pa_GetErrorText(pa_err));
    fflush(sndlog);   
    if(pa_err == paDeviceUnavailable)return -1;
    if(pa_err == paInvalidDevice)    return -1;
    if( pa_err == paFormatIsSupported )
      { 
      maxStandardSampleRate=standardSampleRates[i];
      }
    i--;
   }
// check if min SampleRate  = max SampleRate 
if(maxStandardSampleRate==0 && 
         (minStandardSampleRate != 0))
    {
    maxStandardSampleRate= minStandardSampleRate;
    }
// check min and max bytes
minBytes=2;
maxBytes=2;
if(sound_type==RXDA || sound_type==TXDA)
    {
    outputParameters.sampleFormat = paUInt8;
    pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, maxStandardSampleRate );
    }
else
    {  
    inputParameters.sampleFormat = paUInt8;
    pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, maxStandardSampleRate );
    }
if( pa_err == paFormatIsSupported )
    { 
    minBytes=1;
    }
SNDLOG"\nminBytes=%i", minBytes); 
fflush(sndlog);
if(sound_type==RXDA || sound_type==TXDA)
    {
    outputParameters.sampleFormat = paInt32;
    pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, maxStandardSampleRate );
    }
else
    {  
    inputParameters.sampleFormat = paInt32;
    pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, maxStandardSampleRate );
    }
if( pa_err == paFormatIsSupported )
    { 
    maxBytes=4;
    }
SNDLOG"\nmaxBytes=%i", maxBytes); 
fflush(sndlog);
// check min channel count
maxInputChannels= deviceInfo->maxInputChannels;
maxOutputChannels= deviceInfo->maxOutputChannels;
if(sound_type==RXDA || sound_type==TXDA)
    {
      outputParameters.channelCount = 1;
      outputParameters.sampleFormat = paInt16;
      pa_err=paFormatIsSupported+32000;
      while(pa_err != paFormatIsSupported &&
           ( outputParameters.channelCount < (maxOutputChannels+1)) )
        {
        pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, maxStandardSampleRate );
        outputParameters.channelCount++;
        }
      if( pa_err == paFormatIsSupported )
        {
        minOutputChannels=outputParameters.channelCount-1;
        SNDLOG"\nminOutputChannels=%i", minOutputChannels); 
        SNDLOG"\nmaxOutputChannels=%i\n\n", maxOutputChannels); 
        fflush(sndlog); 
        }
      else
        {
        SNDLOG"\nminOutputChannels=%i samplerate=%6.0f err=%d  %s"
           , (outputParameters.channelCount-1), maxStandardSampleRate, pa_err, Pa_GetErrorText(pa_err));
        fflush(sndlog);
        return -1;
        }
    }
else
    { 
      inputParameters.channelCount = 1;
      inputParameters.sampleFormat = paInt16;
      pa_err=paFormatIsSupported+32000;
      while(pa_err != paFormatIsSupported &&
           ( inputParameters.channelCount < (maxInputChannels+1)) )
        {
        pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, maxStandardSampleRate );
        inputParameters.channelCount++;
        }
      if( pa_err == paFormatIsSupported )
        {
        minInputChannels=inputParameters.channelCount-1;
        SNDLOG"\nminInputChannels=%i", minInputChannels); 
        SNDLOG"\nmaxInputChannels=%i\n\n", maxInputChannels); 
        fflush(sndlog);
        }
      else
        {
        SNDLOG"\nminInputChannels=%i samplerate=%6.0f err=%d  %s"
            ,( inputParameters.channelCount-1), maxStandardSampleRate, pa_err, Pa_GetErrorText(pa_err));
        fflush(sndlog);
        return -1;
        }
     }
#if (OSNUM == OS_FLAG_WINDOWS)
end_pa_get_device_info:;
#endif
*pa_device_max_speed=maxStandardSampleRate;
*pa_device_min_speed=minStandardSampleRate;
*pa_device_max_bytes=maxBytes;
*pa_device_min_bytes=minBytes;
if(sound_type==RXDA || sound_type==TXDA)
    {
    *pa_device_max_channels= maxOutputChannels;
    *pa_device_min_channels= minOutputChannels;
    }
else
    {
    *pa_device_max_channels= maxInputChannels;
    *pa_device_min_channels= minInputChannels;
    }
// Change the limits for RXDA if they are too generous or seem incorrect.
if(sound_type==RXDA )
    {
    if(*pa_device_max_bytes > 2)*pa_device_max_bytes=2;
    if(*pa_device_min_channels >=  2)*pa_device_min_channels=2;
           else *pa_device_min_channels=1;
    if(*pa_device_max_channels <  2 )*pa_device_max_channels=1;
           else *pa_device_max_channels=2;
    }
return speed_warning;
}

void set_portaudio(int sound_type, int *line)
{
(void) sound_type;
(void) line;
char s[128];
int i, j, k;
char devflag[MAX_PA_DEVICES];
double devmaxStandardSampleRate[MAX_PA_DEVICES];
double devminStandardSampleRate[MAX_PA_DEVICES];
int devmaxBytes[MAX_PA_DEVICES];
int devminBytes[MAX_PA_DEVICES];
int devmaxInputChannels[MAX_PA_DEVICES];
int devminInputChannels[MAX_PA_DEVICES];
int devmaxOutputChannels[MAX_PA_DEVICES];
int devminOutputChannels[MAX_PA_DEVICES];
int id, err, numDevices, valid_dev_cnt;
char pa_device_name[128];     
char pa_device_hostapi[128]; 
double pa_device_max_speed;
double pa_device_min_speed;
int pa_device_max_bytes;
int pa_device_min_bytes;
int pa_device_max_channels;
int pa_device_min_channels;

SNDLOG"\nTESTING ALL PORTAUDIO DEVICES FOR %s\n",sndtype[sound_type]);
SNDLOG"\nPortAudio ver %d \n%s\n", Pa_GetVersion(),Pa_GetVersionText());
valid_dev_cnt=0;
clear_screen();
line[0]=4;
numDevices=Pa_GetDeviceCount();
if(numDevices < 0)
  {
  lirerr(1166);
  return;
  }
if(numDevices == 0)
  {
  lir_text(5,2,"No portaudio device detected for testing.");
  SNDLOG"\nNo portaudio device detected for testing.");
  lir_text(5,4,press_any_key);
  await_keyboard();
  goto end_portaudiotest;
  }
  else
   {
   SNDLOG"\nNumber of portaudio devices to be tested=%i\n",numDevices);
   }
fflush(sndlog);
settextcolor(14);
sprintf(s,"LIST OF VALID PORTAUDIO DEVICES FOR %s :",sndtype[sound_type]);
//lir_text(20,0,s);
lir_text(2,1,s);
if(numDevices > MAX_PA_DEVICES)
  {
  lir_text(2,line[0],"TOO MANY DEVICES. List truncated.");
  numDevices=MAX_PA_DEVICES;
  line[0]++;
  }
settextcolor(7);
// headings
lir_text(75,line[0],"SAMPLE RATE");
lir_text(91,line[0],"CHANNELS");
lir_text(106,line[0],"BYTES");
line[0]++;
lir_text(2,line[0],"ID         NAME");
lir_text(54,line[0],"HOST-API");
lir_text(77,line[0],"MIN/MAX");
lir_text(91,line[0],"MIN/MAX");
lir_text(105,line[0],"MIN/MAX");
line[0]++;
lir_text(2,line[0],"--  ------------------------");
lir_text(54,line[0],"-------");
lir_text(74,line[0],"-------------");
lir_text(91,line[0],"-------");
lir_text(105,line[0],"-------");
line[0]++;
lir_refresh_screen();
// look for valid devices
for(id=0; id < (numDevices); id++)
  {
  devflag[id]=0;
  err=pa_get_device_info (id,
                          sound_type,
                          &pa_device_name,
                          &pa_device_hostapi,
			  &pa_device_max_speed,
			  &pa_device_min_speed,
			  &pa_device_max_bytes,
			  &pa_device_min_bytes,
			  &pa_device_max_channels,
			  &pa_device_min_channels
						 );
  if (err >= 0 )
    {
    if(strcmp( pa_device_hostapi,"ASIO")!=0 ||asio_flag==FALSE)
      {
      devflag[id]=1+err;
      valid_dev_cnt++;
//display result on screen
      sprintf(s,"%2d/%2d",pa_device_min_channels,pa_device_max_channels);
      lir_text(91,line[0],s);
      sprintf(s,"%02d  %s", id, pa_device_name);
//    lir_text(12,line[0],s);
      lir_text(2,line[0],s);
      sprintf(s," %s", pa_device_hostapi );
      lir_text(53,line[0],s);
      sprintf(s,"%6.0f/%6.0f",pa_device_min_speed, pa_device_max_speed);
      lir_text(73,line[0],s);
      sprintf(s,"%i/%i",pa_device_min_bytes, pa_device_max_bytes);
      lir_text(107,line[0],s);
      line[0]++;
      if(line[0] >= screen_last_line)
        {
screenerr:;        
        lirerr(1235);
        return;
        }
      lir_refresh_screen();
//save values
      devmaxStandardSampleRate[id]=pa_device_max_speed;
      devminStandardSampleRate[id]=pa_device_min_speed;
      devmaxBytes[id]=pa_device_max_bytes;
      devminBytes[id]=pa_device_min_bytes;
      if(sound_type==RXDA || sound_type==TXDA)
        {
        devmaxOutputChannels[id]=pa_device_max_channels;
        devminOutputChannels[id]=pa_device_min_channels;
        }
      else
        {
        devmaxInputChannels[id]=pa_device_max_channels;
        devminInputChannels[id]=pa_device_min_channels;
        }
      }   
    }
  }
SNDLOG"\n\nEND TESTING PORTAUDIO DEVICES\n");
fflush(sndlog);
if (valid_dev_cnt !=0)
  {
  line[0]+=3;
  if(line[0] >= screen_last_line)goto screenerr;
  sprintf(s,"Select portaudio device ID for %s from list >",sndtype[sound_type]);
  SNDLOG"\n%s",s);
  gtdev:;
  settextcolor(14);
  clear_lines(line[0],line[0]+1);
  lir_text(2,line[0],s);
  settextcolor(7);
  id=lir_get_integer(strlen(s)+4,line[0], 2, 0, numDevices-1);
  SNDLOG"\nUser selected %i\n",id);
  if(kill_all_flag)return;
  if(devflag[id] == 0)goto gtdev;
  if(devflag[id] != 1)
    {
    lir_text(2,line[0]+1,"This is a very slow device. Really use it? (Y/N)");
really:;
    toupper_await_keyboard();
    if(lir_inkey == 'N')goto gtdev;
    if(lir_inkey != 'Y')goto really;
    line[0]++;
    if(line[0] >= screen_last_line)goto screenerr;
    }
  line[0]++;
  if(line[0] >= screen_last_line-6)
    {
    line[0]=5;
    clear_screen();
    }
  switch (sound_type)
    {
    case RXDA:
    ui.rx_dadev_no=id;
    ui.rx_max_da_speed=devmaxStandardSampleRate[id];
    ui.rx_min_da_speed=devminStandardSampleRate[id];
    ui.rx_min_da_bytes=devminBytes[id];
    ui.rx_max_da_bytes=devmaxBytes[id];
    ui.rx_min_da_channels= devminOutputChannels[id];
    ui.rx_max_da_channels=devmaxOutputChannels[id];
    if(ui.operator_skil != OPERATOR_SKIL_NEWCOMER)       
      {
      line[0]+=2;
      sprintf(s,"Latency Factor (1 to %d)",MAX_LATENCY);
      lir_text(0,line[0],s);
      ui.rx_da_latency=lir_get_integer(26,line[0], 2, 1, MAX_LATENCY);
      }
    else
      {
      ui.rx_da_latency=5;
      }  
    break;

    case RXAD:
    ui.rx_addev_no=id;
    line[0]+=2;
    lir_text(10,line[0],"Sampling speed (Hz)>");
    ui.rx_ad_speed=lir_get_integer(31,line[0], 6, 
                                         devminStandardSampleRate[id], 
                                                devmaxStandardSampleRate[id]);
    if(kill_all_flag) return;
    if(devmaxBytes[id] < 2)
      {
      lirerr(1217);
      return;
      }  
    if(devminBytes[id] < 2)devminBytes[id]=2;
    ui.rx_input_mode=0;
    if(devmaxBytes[id] == 3)
      {
      lirerr(366132);
      return;
      }
get_rxbits:;
    lir_text(40,line[0],"No of bits (16/32):");
    i=lir_get_integer(60,line[0], 2, 16, 32);
    if(kill_all_flag) return;
    clear_screen();
    if(i==32)
      {
      ui.rx_input_mode=DWORD_INPUT;
      }
    else
      {
      if(i != 16)goto get_rxbits;
      }
    clear_screen();
// Ask for the desired processing mode.
sel_radio:;
    clear_screen();
    lir_text(10,10,"Select radio interface>");
    j=1;
    if(devmaxInputChannels[id]>=2)j=3;
    if(devmaxInputChannels[id]>=4)j=4;
    if(devminInputChannels[id]==1)
      {
      k=0;
      }
    else
      {
      k=1;
      }  
    for(i=k; i<j; i++)
      {
      sprintf(s,"%d: %s",i+1,audiomode_text[i]);
      lir_text(10,12+i,s);
      }
    lir_text(10,13+i,"F1 for help/info");
    line[0]=15+i;
chsel:;
    await_processed_keyboard();
    if(lir_inkey == F1_KEY || lir_inkey == '!')
      {
      if(ui.operator_skil != OPERATOR_SKIL_NEWCOMER)
        {
        help_message(89);
        }
      else
        {  
        help_message(109);
        }
      goto sel_radio;
      }
    if(lir_inkey-'0'>j)goto chsel;
    if(lir_inkey-'0'<k)goto chsel;
    switch (lir_inkey)
      {
      case '1':
      ui.rx_rf_channels=1;
      ui.rx_ad_channels=1;
      break;

      case '2':
      ui.rx_input_mode|=IQ_DATA;
      ui.rx_rf_channels=1;
      ui.rx_ad_channels=2;
      break;

      case '3':
      ui.rx_input_mode|=TWO_CHANNELS;
      ui.rx_rf_channels=2;
      ui.rx_ad_channels=2;
      break;

      case '4':
      ui.rx_input_mode|=TWO_CHANNELS+IQ_DATA;
      ui.rx_rf_channels=2;
      ui.rx_ad_channels=4;
      break;

      default:
      goto chsel;
      }
    
// Input seems OK.
    if(ui.rx_rf_channels == 1 && ui.rx_ad_channels == 2)
      {
      line[0]+=2;
      lir_text(0,line[0],
                "Number of points to time shift between I and Q? (-4 to +4)");
      ui.sample_shift=lir_get_integer(60,line[0], 2, -4, 4);
      }
    if(ui.operator_skil != OPERATOR_SKIL_NEWCOMER)       
      {
      line[0]+=2;
      sprintf(s,"Latency Factor (1 to %d)",MAX_LATENCY);
      lir_text(0,line[0],s);
      ui.rx_ad_latency=lir_get_integer(26,line[0], 2, 1, MAX_LATENCY);
      }
    else
      {
      ui.rx_ad_latency=5;
      }  
    } 
  }
else
  {
  line[0]++;
  settextcolor(14);
  lir_text(2,line[0],"No valid portaudio device detected ");
  SNDLOG"\nNo valid portaudio device detected ");
  line[0]++;
  lir_text(2,line[0],press_any_key);
  settextcolor(7);
  await_keyboard();
  } 
end_portaudiotest:;
}

int pa_get_valid_samplerate (int n,
                             int mode,
                             int *line,
                             unsigned int *returned_sampling_rate)
{
(void) n;
(void) mode;
(void) line;
(void) returned_sampling_rate;
int i,k,id;
unsigned int val;
char s[80];
unsigned int valid_sampling_rate[16];
// negative terminated  list
static double standardSampleRates[] = {8000.0, 9600.0, 
        11025.0, 12000.0, 16000.0, 22050.0, 24000.0, 32000.0,
        44100.0, 48000.0, 88200.0, 96000.0, 192000.0, -1};
const PaDeviceInfo *deviceInfo;
PaError pa_err;
PaStreamParameters inputParameters;
PaStreamParameters outputParameters;
deviceInfo = Pa_GetDeviceInfo(n);
inputParameters.device = n;
inputParameters.channelCount = deviceInfo->maxInputChannels;
inputParameters.sampleFormat = paInt16;
inputParameters.suggestedLatency = 0;
inputParameters.hostApiSpecificStreamInfo = NULL;
outputParameters.device = n;
outputParameters.channelCount = deviceInfo->maxOutputChannels;
outputParameters.sampleFormat = paInt16; 
outputParameters.suggestedLatency = 0;
outputParameters.hostApiSpecificStreamInfo = NULL;
line[0]+=2;
sprintf(s,"This is a portaudio device.");
lir_text(1,line[0], s);
line[0]+=1;
sprintf(s,"Select sampling rate by line number from list :");
lir_text(1,line[0], s);
line[0]+=2;
#if (OSNUM == OS_FLAG_WINDOWS)
// bypass bug in Juli@ ASIO driver: 
// this driver hangs after a Pa_IsFormatSupported call 
static double JuliSampleRates[] = {
       44100.0, 48000.0, 88200.0, 96000.0, 176400., 192000.0, -1};
i = strncmp(deviceInfo->name, "ASIO 2.0 - ESI Juli@", 19);
k=0;
if (i == 0)
    {
    while(JuliSampleRates[k] > 0 )
      {
      valid_sampling_rate[k]= JuliSampleRates[k];
      sprintf(s,"%2d -> %6d ",k,valid_sampling_rate[k]);
      lir_text(1,line[0],s);
      line[0]++;
      k++;
      }
    goto end_pa_get_valid_samplerate;
    }
#endif
i=0;
k=0;
while(standardSampleRates[i] > 0 )
   {
    val= standardSampleRates[i];
    if(mode==RXDA || mode==TXDA)
      {
      pa_err=Pa_IsFormatSupported(
                          NULL, &outputParameters, standardSampleRates[i] );
      }
    else
      {  
      pa_err=Pa_IsFormatSupported(
                  &inputParameters, NULL, standardSampleRates[i] );
      }
    if(pa_err == paFormatIsSupported )
      { 
      sprintf(s,"%2d -> %6d ",k,val);
      lir_text(1,line[0],s);
      valid_sampling_rate[k]= val;
      line[0]++;
      k++;      
      }
    i++;
   }
#if (OSNUM == OS_FLAG_WINDOWS)
end_pa_get_valid_samplerate:;
#endif
line[0]++;
sprintf(s,"Enter line number >");
lir_text(1,line[0],s);
id=lir_get_integer(21,line[0],2, 0, k-1);
*returned_sampling_rate=valid_sampling_rate[id];
return 0;
}

void portaudio_startstop(void)
{
PaErrorCode err;
if( (ui.use_alsa&PORTAUDIO_RX_IN) != 0 ||
    (ui.use_alsa&PORTAUDIO_RX_OUT) != 0 ||
    (ui.use_alsa&PORTAUDIO_TX_IN) != 0 ||
    (ui.use_alsa&PORTAUDIO_TX_OUT) != 0 )
  {
  if(portaudio_active_flag == FALSE)
    {
#if(OSNUM == OS_FLAG_LINUX)
#if JACK_PRESENT == 1
    PaJack_SetClientName(linrad_name);
#endif
#endif
    err=Pa_Initialize();
    if(err < 0)
      {
      lirerr(847297);
      return;
      }      
DEB"\nPortAudio ver %d \n%s\n", Pa_GetVersion(), Pa_GetVersionText());
    }
  portaudio_active_flag=TRUE;
  }
else
  {  
  if(portaudio_active_flag == TRUE)
    {
    err=Pa_Terminate();
DEB"\nPortaudio terminated.\n");
    if(err < 0)
      {
      lirerr(847298);
      return;
      }
    }
  portaudio_active_flag=FALSE;
  }
}
#endif

